﻿using Microsoft.Extensions.Logging;
using Spider.Core.DataProvider;
using Spider.Core.Pipeline;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

namespace Spider.Core
{
    /// <summary>
    /// 商品批处理器。
    /// </summary>
    public class DataCollectBatchProcessor : IDataCollectProcessor
    {
        private readonly IExtractable _extractor;
        private readonly ITransformable _transformer;
        private readonly ILoadable _loader;
        private readonly ILogger _logger;
        private readonly IDataProvider<Production> _dataProvider;

        protected static class RunningState
        {
            internal const int Idle = 0;
            internal const int Running = Idle + 1;
        }

        private volatile int _running;

        public DataCollectBatchProcessor(
            IExtractable extractor,
            ITransformable transformer,
            ILoadable loader,
            ILogger<DataCollectBatchProcessor> logger,
            IDataProvider<Production> dataProvider)
        {
            _extractor = extractor;
            _transformer = transformer;
            _loader = loader;
            _logger = logger;
            _dataProvider = dataProvider;
        }

        public async Task RunAsync<TState>(TState state)
        {
            if (Interlocked.CompareExchange(ref _running, RunningState.Running, RunningState.Idle) == RunningState.Idle)
            {
                if (!(state is DataCollectContext context)) return;

                var start = 0;

                // 分小批次平滑处理引擎分配的大批次。
                var largeBatchSize = context.LargeBatchSize;
                var smallBatchSize = context.SmallBatchSize;
                var round = context.SmallBatch <= 0 ? 1 : context.SmallBatch;
                var batchSize = largeBatchSize > smallBatchSize ? smallBatchSize : largeBatchSize;

                using (_logger.BeginScope($"Data Collect Batch Processor {context}"))
                {
                    try
                    {
                        while (start <= largeBatchSize)
                        {
                            // double check
                            if (_running != RunningState.Running)
                                continue;

                            _logger.LogInformation(
                                $"{Environment.NewLine}Small batch: the {round} round[{start}-{start + batchSize}] beginning...");

                            var productions = await _dataProvider.GetAsync(context);
                            await ProcessingAsync(context, productions);

                            _logger.LogInformation($"Small batch: the {round} round[{start}-{start + batchSize}] ended.");

                            start += batchSize;
                            context.SmallBatch = round++;
                        }
                    }
                    catch (Exception e)
                    {
                        _logger.LogError($"{nameof(RunAsync)} has a unknown exception: {e}.");
                    }
                    finally
                    {
                        Interlocked.Exchange(ref _running, RunningState.Idle);
                    }
                }
            }
            else
            {
                _logger.LogError("Thread is already running.");
            }
        }

        private async Task ProcessingAsync(DataCollectContext context,
            IEnumerable<Production> productions)
        {
            if (productions == null)
                return;

            var index = 0;
            foreach (var production in productions)
            {
                var trace = $"[index: {++index}]";

                _logger.LogInformation($"{Environment.NewLine}{trace}Extracting");
                var payloads = _extractor.Extract(production);
                _logger.LogInformation($"{trace}Extracted");

                if (payloads == null || !payloads.Any())
                    continue;

                _logger.LogInformation($"{trace}Transforming");
                await _transformer.TransformAsync(context, payloads);
                _logger.LogInformation($"{trace}Transformed");

                _logger.LogInformation($"{trace}Loading");
                await _loader.LoadAsync(payloads);
                _logger.LogInformation($"{trace}Loaded");
            }
        }
    }
}
