﻿using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Options;
using Spider.Core.DataProvider;
using Spider.Core.Exceptions;
using Spider.Infrastructure;
using System;
using System.Threading;
using System.Threading.Tasks;

namespace Spider.Core
{
    public class DataCollectEngine : IDataCollectEngine
    {
        private volatile int _started;

        private readonly IExecutor _executor;
        private readonly DataCollectorOptions _options;
        private readonly IDataProvider<Production> _dataProvider;
        private readonly IServiceProvider _serviceProvider;
        private readonly ILogger<DataCollectEngine> _logger;

        public DataCollectEngine(
            IExecutor executor,
            IOptions<DataCollectorOptions> options,
            IDataProvider<Production> dataProvider,
            ILogger<DataCollectEngine> logger,
            IServiceProvider serviceProvider)
        {
            _executor = executor;
            _dataProvider = dataProvider;
            _logger = logger;
            _serviceProvider = serviceProvider;
            _options = options.Value;
        }

        public async Task StartAsync()
        {
            CheckOnlyStartedOnce();

            var sizeToProcessed = await _dataProvider.GetCountAsync();
            if (sizeToProcessed <= 0)
            {
                _logger.LogCritical($"[{AppEnvironment.Name}]Engine Startup failure: Total size to processed is {sizeToProcessed}.");
                return;
            }

            var checkPointId = _options.CheckPointId;
            if (_options.CheckPointId <= 0)
                checkPointId = await _dataProvider.GetCheckPointIdAsync();

            var context = new DataCollectContextBuilder()
            .AddCheckPoint(checkPointId)
            .AddTotalSize(sizeToProcessed)
            .AddOptions(_options)
            .Build();

            await DisruptAndDoWorkAsync(context);
        }

        private async Task DisruptAndDoWorkAsync(DataCollectContext context)
        {
            var rounds = context.Rounds;
            var workers = new Task[rounds];

            using (_logger.BeginScope($"[{AppEnvironment.Name}]Data Collect Engine"))
            {
                for (var j = 0; j < rounds; j++)
                {
                    context.LargeBatch = j + 1;

                    using (var scope = _serviceProvider.CreateScope())
                    {
                        var processor = scope.ServiceProvider.GetRequiredService<IDataCollectProcessor>();
                        workers[j] = _executor.ExecuteAsync(processor, context.Clone());
                    }
                }

                await Task.WhenAll(workers).ConfigureAwait(false);
            }

            _logger.LogInformation($"The work has been done.");
        }

        private void CheckOnlyStartedOnce()
        {
            if (Interlocked.Exchange(ref _started, 1) == 1)
            {
                throw new IllegalStateException($"{nameof(StartAsync)} must only be called once.");
            }
        }
    }
}
